package gs.log.v0.api

import gs.log.v0.data.Log
import gs.log.v0.data.LogLevel
import cats.Show

/**
  * Interface for emitting logs. This and [[Log]] are the primary types that
  * users of this library will interact with.
  */
trait Logger[F[_]]:
  /**
   * Emit the given [[Log]] at the [[LogLevel.Trace]] level.
   *
   * @param log The [[Log]] to emit.
   * @return Side-effect.
   */
  def trace(log: => Log): F[Unit]

  /**
   * Emit the given [[Log]] at the [[LogLevel.Debug]] level.
   *
   * @param log The [[Log]] to emit.
   * @return Side-effect.
   */
  def debug(log: => Log): F[Unit]

  /**
   * Emit the given [[Log]] at the [[LogLevel.Info]] level.
   *
   * @param log The [[Log]] to emit.
   * @return Side-effect.
   */
  def info(log: => Log): F[Unit]

  /**
   * Emit the given [[Log]] at the [[LogLevel.Warn]] level.
   *
   * @param log The [[Log]] to emit.
   * @return Side-effect.
   */
  def warn(log: => Log): F[Unit]

  /**
   * Emit the given [[Log]] at the [[LogLevel.Error]] level.
   *
   * @param log The [[Log]] to emit.
   * @return Side-effect.
   */
  def error(log: => Log): F[Unit]

  /**
   * Emit the given [[Log]] at the [[LogLevel.Fatal]] level.
   *
   * @param log The [[Log]] to emit.
   * @return Side-effect.
   */
  def fatal(log: => Log): F[Unit]

  /**
   * @return The name of this logger.
   */
  def name(): Logger.Name

  /**
   * Determine whether this logger has the given level enabled.
   *
   * @param level The [[LogLevel]] to check.
   * @return True if the level is enabled, false otherwise.
   */
  def isLevelEnabled(level: LogLevel): Boolean

object Logger:

  /** Opaque type (String) used to represent the descriptive name of a logger. 
   *  If a logger has a name, logs _may_ include that name in their output (if 
   *  configured to do so).
    */
  opaque type Name = String

  object Name:

    /**
     * Instantiate a new [[Name]] from the given string.
     *
     * @param loggerName The logger's name.
     * @return
     */
    def apply(loggerName: String): Name = loggerName

    given CanEqual[Name, Name] = CanEqual.derived

    given Show[Name] = new Show[Name] {
      override def show(t: Name): String = t
    }

  end Name

end Logger
