package gs.log.v0.data

import java.time.Instant
import scala.collection.mutable.{Map => MutMap}

/**
  * Representation of a Log - a package of data that can be rendered and emitted
  * by a [[Logger]] and underlying engine. This type provides a fluent interface
  * for constructing and logs.
  *
  * @param data Key/value mapping of data that is safe to log.
  * @param tags List of tags (see: [[LogTags]]) applicable to this log.
  * @param message Optional [[LogMessage]] to include in the output.
  * @param exception Optional exception to include in the output.
  */
final class Log(
  private val internalData: MutMap[LogData.Key, LogData.Safe] = MutMap.empty,
  private var internalTags: List[String] = List.empty,
  private var internalMessage: Option[LogMessage] = None,
  private var internalException: Option[Throwable] = None
):
  /**
   * Set the value for a single key.
   *
   * @param key The key to set.
   * @param value The safe value for the key.
   * @return This [[Log]].
   */
  def data(key: LogData.Key, value: LogData.Safe): Log =
    val _ = this.internalData.put(key, value)
    this

  /**
   * Set a list of key/value pairs.
   *
   * @param kvp The list of key/value pairs.
   * @return This [[Log]].
   */
  def data(kvp: (LogData.Key, LogData.Safe)*): Log =
    val _ = kvp.foreach {
      case (k, v) => val _ = this.internalData.put(k, v)
    }
    this

  /**
   * Replace the current list of tags with a new list of tags.
   *
   * @param newTags The new list of tags for this log.
   * @return This [[Log]].
   */
  def tagged(newTags: String*): Log =
    this.internalTags = newTags.toList
    this

  /**
   * Replace the current message with a new message.
   *
   * @param newMessage The new message.
   * @return This [[Log]].
   */
  def message(newMessage: LogMessage): Log =
    this.internalMessage = Some(newMessage)
    this

  /**
   * Replace the current exception with a new exception.
   *
   * @param ex The new exception.
   * @return This [[Log]].
   */
  def exception(ex: Throwable): Log =
    this.internalException = Some(ex)
    this

  def getData: Map[LogData.Key, LogData.Safe] = internalData.toMap

  def getTags: List[String] = internalTags

  def getMessage: Option[LogMessage] = internalMessage

  def getException: Option[Throwable] = internalException

  /**
   * Determine if _any_ of the fields on this [[Log]] are set. Logs that are
   * empty are always discarded.
   *
   * @return Whether or not this [[Log]] is empty.
   */
  def isEmpty: Boolean =
    internalData.isEmpty 
      && internalTags.isEmpty 
      && internalMessage.isEmpty 
      && internalException.isEmpty

object Log:
  /**
   * Instantiate a new log with the given key/value pair.
   *
   * @param key The key.
   * @param value The value.
   * @return New [[Log]] instance.
   */
  def data(key: LogData.Key, value: LogData.Safe): Log =
    new Log().data(key, value)

  /**
   * Instantiate a new log with the given list of key/value pairs.
   *
   * @param kvp The list of key/value pairs.
   * @return New [[Log]] instance.
   */
  def data(kvp: (LogData.Key, LogData.Safe)*): Log =
    new Log().data(kvp*)

  /**
   * Instantiate a new log with the given list of tags.
   *
   * @param tags The list of tags.
   * @return New [[Log]] instance.
   */
  def tagged(tags: String*): Log =
    new Log().tagged(tags*)

  /**
   * Instantiate a new log with the given message.
   *
   * @param logMessage The message.
   * @return New [[Log]] instance.
   */
  def message(logMessage: LogMessage): Log =
    new Log().message(logMessage)

  /**
   * Instantiate a new log with the given exception.
   *
   * @param tags The exception.
   * @return New [[Log]] instance.
   */
  def exception(ex: Throwable): Log =
    new Log().exception(ex)

  /** All supplementary information about a [[Log]] as generated/passed by the
    * [[Logger]]. This type should not typically be used directly by users.
    *
    * @param level
    *   The selected [[LogLevel]]
    * @param timestamp
    *   The instant the [[Log]] was _submitted_.
    * @param owner
    *   The name of the [[Logger]] which submitted the log.
    */
  case class Metadata(
    level: LogLevel,
    timestamp: Instant,
    owner: String
  )

end Log
