package gs.log.v0.data

import java.time.LocalDate

object LogData:

  /** Opaque type (String) that represents a key that identifies some piece of
    * data in a [[Log]].
    */
  opaque type Key = String

  object Key:

    /** Instantiate a new [[Key]] from the given input string.
      *
      * @param value
      *   The value of the key.
      * @return
      *   The new [[Key]] instance.
      */
    def apply(value: String): Key = value

    given CanEqual[Key, Key] = CanEqual.derived

    extension (key: Key)
      /** Express this key as a `String`.
        */
      def str(): String = key

  end Key

  /** Represents some data that can be logged. This is the target of the
    * [[Loggable]] type class. Note that `Value` is not typically used directly
    * by callers - [[Safe]] is used to ensure that all data is explicitly
    * denoted.
    */
  trait Value:
    type A

    /** @return
      *   The data to log.
      */
    def data: A

  /**
   * Provide standard capture support for strings.
   *
   * @param data The string data.
   */
  case class Str(data: String) extends Value:
    override type A = String

  /**
   * Provide standard capture support for Boolean values.
   *
   * @param data The Boolean data.
   */
  case class Bool(data: Boolean) extends Value:
    override type A = Boolean

  /**
   * Provide standard capture support for 32-bit integers.
   *
   * @param data The integer data.
   */
  case class Int32(data: Int) extends Value:
    override type A = Int

  /**
   * Provide standard capture support for 64-bit integers.
   *
   * @param data The long integer data.
   */
  case class Int64(data: Long) extends Value:
    override type A = Long

  /**
   * Provide standard capture support for 32-bit floating point numbers.
   *
   * @param data The floating point data.
   */
  case class Float32(data: Float) extends Value:
    override type A = Float

  /**
   * Provide standard capture support for 64-bit floating point numbers.
   *
   * @param data The double-precision floating point data.
   */
  case class Float64(data: Double) extends Value:
    override type A = Double

  /**
   * Provide standard capture support for dates.
   *
   * @param data The date data.
   */
  case class Date(data: LocalDate) extends Value:
    override type A = LocalDate

  /**
   * Provide standard capture support for instants.
   *
   * @param data The instant data.
   */
  case class Instant(data: java.time.Instant) extends Value:
    override type A = java.time.Instant

  trait Generic[AA, Codec[_]] extends Value:
    override type A = AA
    def codec: Codec[A]

  /** Indicates data (in terms of [[Value]]) that has been explicitly marked as
    * _safe_ to log in some form.
    */
  sealed trait Safe

  object Safe:
    
    given CanEqual[Safe, Safe] = CanEqual.derived

  end Safe

  /** The contained data is safe to log in clear text. This means that it will
    * show up in output and be human-readable. Use this for values that are not
    * sensitive.
    *
    * If you have configured a log index where it is considered safe to log all
    * values (e.g. PHI) in clear text, you still should _NOT_ use this type! In
    * those cases, still mark the data appropriately (encrypted/hashed) but use
    * engine configuration to disable those functions and log in clear text.
    * This practice provides better auditing and portability.
    *
    * @param value
    *   The value that is safe to log in clear text.
    */
  case class ClearText(value: Value) extends Safe

  object ClearText:

    given CanEqual[ClearText, ClearText] = CanEqual.derived

  end ClearText

  /** The contained data is _not_ safe to log in clear text, and must be
    * encrypted before being output. In this use case, the data is valuable to
    * have available for debugging/investigation, but it is not safe to include
    * in the clear. Users must _decrypt_ the data before using it.
    *
    * Note that `gs-log` does not provide a general decryption tool, nor should
    * it! It is your responsibility to provide a secure way to decrypt sensitive
    * information extracted from logs.
    *
    * As an alternative, if you have a log index that is certified to store all
    * types of data in the clear due to other protections, use this type but
    * turn off encryption in engine configuration.
    *
    * This type is not suitable for data that needs to be _searchable_. For that
    * use case, please refer to [[RequiresSecureHash]].
    *
    * Example Algorithm: AES GCM (256-bit)
    *
    * @param value
    *   The value to encrypt.
    */
  case class RequiresEncryption(value: Value) extends Safe

  object RequiresEncryption:

    given CanEqual[RequiresEncryption, RequiresEncryption] = CanEqual.derived

  end RequiresEncryption

  /** The contianed data is _not_ safe to log in clear text, and must be hashed
    * before being output. In this use case, the data is valuable to be searched
    * upon to narrow results, but is not safe to include in the clear. Users
    * must generate a hash using the same parameters to search.
    *
    * For example, consider a case where in rare cases certain customers have
    * isolated issues and searching by first and last name is useful for
    * narrowing results. In this case, the first and last names are hashed in
    * the log output -- the user must select a user, hash their first and last
    * name, and enter those hashes as search criteria.
    *
    * As an alternative, if you have a log index that is certified to store all
    * types of data in the clear due to other protections, use this type but
    * turn off hashing in engine configuration.
    *
    * This type is not suitable for data that needs to be _decrypted_. For that
    * use case, please refer to [[RequiresEncryption]].
    *
    * Example algorithm: HMAC + SHA-256
    *
    * @param value
    *   The value to hash.
    */
  case class RequiresSecureHash(value: Value) extends Safe

  object RequiresSecureHash:

    given CanEqual[RequiresSecureHash, RequiresSecureHash] = CanEqual.derived

  end RequiresSecureHash

  /** The specified data is safe to log in clear text. This means that it
    * will show up in output and be human-readable. Use this for values that
    * are not sensitive.
    *
    * If you have configured a log index where it is considered safe to log
    * all values (e.g. PHI) in clear text, you still should _NOT_ use this
    * type! In those cases, still mark the data appropriately
    * (encrypted/hashed) but use engine configuration to disable those
    * functions and log in clear text. This practice provides better
    * auditing and portability.
    *
    * @param value
    *   The value that is safe to log in clear text.
    */
  def logClearText[A: Loggable](data: A): ClearText =
    ClearText(Loggable[A].capture(data))

  /** The specified data is _not_ safe to log in clear text, and must be
    * encrypted before being output. In this use case, the data is valuable
    * to have available for debugging/investigation, but it is not safe to
    * include in the clear. Users must _decrypt_ the data before using it.
    *
    * Note that `gs-log` does not provide a general decryption tool, nor
    * should it! It is your responsibility to provide a secure way to
    * decrypt sensitive information extracted from logs.
    *
    * As an alternative, if you have a log index that is certified to store
    * all types of data in the clear due to other protections, use this type
    * but turn off encryption in engine configuration.
    *
    * This type is not suitable for data that needs to be _searchable_. For
    * that use case, please refer to [[RequiresSecureHash]].
    *
    * Example Algorithm: AES GCM (256-bit)
    *
    * @param value
    *   The value to encrypt.
    */
  def logEncrypted[A: Loggable](data: A): RequiresEncryption =
    RequiresEncryption(Loggable[A].capture(data))

  /** The contianed data is _not_ safe to log in clear text, and must be
    * hashed before being output. In this use case, the data is valuable to
    * be searched upon to narrow results, but is not safe to include in the
    * clear. Users must generate a hash using the same parameters to search.
    *
    * For example, consider a case where in rare cases certain customers
    * have isolated issues and searching by first and last name is useful
    * for narrowing results. In this case, the first and last names are
    * hashed in the log output -- the user must select a user, hash their
    * first and last name, and enter those hashes as search criteria.
    *
    * As an alternative, if you have a log index that is certified to store
    * all types of data in the clear due to other protections, use this type
    * but turn off hashing in engine configuration.
    *
    * This type is not suitable for data that needs to be _decrypted_. For
    * that use case, please refer to [[RequiresEncryption]].
    *
    * Example algorithm: HMAC + SHA-256
    *
    * @param value
    *   The value to hash.
    */
  def logHashed[A: Loggable](data: A): RequiresSecureHash =
    RequiresSecureHash(Loggable[A].capture(data))

  trait Syntax:

    extension [A: Loggable](data: A)
      /** The specified data is safe to log in clear text. This means that it
        * will show up in output and be human-readable. Use this for values that
        * are not sensitive.
        *
        * If you have configured a log index where it is considered safe to log
        * all values (e.g. PHI) in clear text, you still should _NOT_ use this
        * type! In those cases, still mark the data appropriately
        * (encrypted/hashed) but use engine configuration to disable those
        * functions and log in clear text. This practice provides better
        * auditing and portability.
        *
        * @param value
        *   The value that is safe to log in clear text.
        */
      def logClearText(): ClearText = LogData.logClearText(data)

      /** The specified data is _not_ safe to log in clear text, and must be
        * encrypted before being output. In this use case, the data is valuable
        * to have available for debugging/investigation, but it is not safe to
        * include in the clear. Users must _decrypt_ the data before using it.
        *
        * Note that `gs-log` does not provide a general decryption tool, nor
        * should it! It is your responsibility to provide a secure way to
        * decrypt sensitive information extracted from logs.
        *
        * As an alternative, if you have a log index that is certified to store
        * all types of data in the clear due to other protections, use this type
        * but turn off encryption in engine configuration.
        *
        * This type is not suitable for data that needs to be _searchable_. For
        * that use case, please refer to [[RequiresSecureHash]].
        *
        * Example Algorithm: AES GCM (256-bit)
        *
        * @param value
        *   The value to encrypt.
        */
      def logEncrypted(): RequiresEncryption = LogData.logEncrypted(data)

      /** The contianed data is _not_ safe to log in clear text, and must be
        * hashed before being output. In this use case, the data is valuable to
        * be searched upon to narrow results, but is not safe to include in the
        * clear. Users must generate a hash using the same parameters to search.
        *
        * For example, consider a case where in rare cases certain customers
        * have isolated issues and searching by first and last name is useful
        * for narrowing results. In this case, the first and last names are
        * hashed in the log output -- the user must select a user, hash their
        * first and last name, and enter those hashes as search criteria.
        *
        * As an alternative, if you have a log index that is certified to store
        * all types of data in the clear due to other protections, use this type
        * but turn off hashing in engine configuration.
        *
        * This type is not suitable for data that needs to be _decrypted_. For
        * that use case, please refer to [[RequiresEncryption]].
        *
        * Example algorithm: HMAC + SHA-256
        *
        * @param value
        *   The value to hash.
        */
      def logHashed(): RequiresSecureHash = LogData.logHashed(data)

  end Syntax

end LogData
