package gs.log.v0.data

/** Enumerates all supported log levels, which control logging verbosity.
  *
  * @param name
  *   The name of the log level.
  * @param ordinal
  *   The order/precedence of the log level (lower level = higher precedence).
  */
sealed abstract class LogLevel(
  val name: String,
  val ordinal: Int
) extends Ordered[LogLevel]:
  /** Compare two [[LogLevel]] instances for the purpose of ordering. More
    * important levels are smaller, more verbose levels are larger.
    *
    * Off < Fatal < Error < Warn < Info < Debug < Trace
    *
    * @param that
    *   The [[LogLevel]] to compare against.
    * @return
    *   The comparison result of this level against the other.
    */
  override def compare(that: LogLevel): Int = ordinal - that.ordinal

object LogLevel:
  given CanEqual[LogLevel, LogLevel] = CanEqual.derived

  /** Most-detailed log level. This level should (likely) not be turned on most
    * of the time. It should be used to provide extremely verbose output.
    */
  case object Trace extends LogLevel("trace", 5)

  /** Used to provide additional detail, typically used for debugging behavior.
    */
  case object Debug extends LogLevel("debug", 4)

  /** Provides a high-level explanation of behavior at runtime. This is the
    * default level of detail for logs.
    */
  case object Info extends LogLevel("info", 3)

  /** Used for cases where something unexpected or "in error" happens, but the
    * software is able to continue operating normally.
    */
  case object Warn extends LogLevel("warn", 2)

  /** Used when the software can no longer operate normally. Things like
    * exceptions typically fall into this category. Things like database
    * connection failures might be represented with error logs.
    */
  case object Error extends LogLevel("error", 1)

  /** Typically represents a case where an application is forced to crash due to
    * some type of error. For example, a web application might report a fatal
    * error if it cannot read required configuration on startup and fail to
    * start entirely.
    */
  case object Fatal extends LogLevel("fatal", 0)

  /** Indicates that logging is completely disabled.
    */
  case object Off extends LogLevel("off", -1)
end LogLevel
