package gs.log.v0.data

import cats.Show

/** Opaque type (String) for representing messages that should be logged. This
  * type can _only_ be instantiated using the `log` string interpolator, which
  * allows it to be (relatively) "safe" in that users _must_ explicitly declare
  * all data they intend to log as safe for clear text.
  *
  * Example:
  * {{{
  * val data = "example".logClearText()
  * val message: LogMessage = log"this is an example '$data'"
  * }}}
  */
opaque type LogMessage = String

object LogMessage:

  given CanEqual[LogMessage, LogMessage] = CanEqual.derived

  given Show[LogMessage] = new Show[LogMessage] {
    override def show(t: LogMessage): String = t
  }

  /**
   * Represents a string that is explicitly safe to show in clear text.
   *
   * @param data The clear text data.
   */
  case class ClearTextString(data: String)

  object ClearTextString:

    given CanEqual[ClearTextString, ClearTextString] = CanEqual.derived

    given Show[ClearTextString] = new Show[ClearTextString] {
      override def show(t: ClearTextString): String = t.data
    }

  end ClearTextString

  extension (logMessage: LogMessage)
    /** Express this [[LogMessage]] as a string.
      *
      * @return
      *   The string rendition of this [[LogMessage]].
      */
    def unwrap(): String = logMessage

  trait Syntax:

    extension [A: Show](data: A)
      /**
       * Explicitly mark this data as _safe_ to render in clear text. Produce a
       * string representation.
       *
       * @return The clear text string representation of this data.
       */
      def clearTextString(): ClearTextString = 
        ClearTextString(Show[A].show(data))

    extension (sc: StringContext)

      /** String interpolator for instantiating a [[LogMessage]].
        *
        * @param args
        *   All interpolation variables. Must be [[LogMessage.ClearTextString]]
        *   instances.
        * @return
        *   The new [[LogMessage]].
        */
      def log(args: ClearTextString*): LogMessage =
        sc.s(args.map(cts => cts.data)*)

end LogMessage
